Pokedex = LibStub("AceAddon-3.0"):NewAddon("Pokedex", "AceConsole-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Pokedex")

-- Keybinding globals --
BINDING_HEADER_POKEDEX = L["Pokedex"]
BINDING_NAME_POKEDEXSUPERTOGGLE = L["Super Toggle"]
BINDING_NAME_POKEDEXSUMMONMOUNT = L["Summon Mount"]
BINDING_NAME_POKEDEXDISMISSMOUNT = L["Dismiss Mount"]
BINDING_NAME_POKEDEXTOGGLEMOUNT = L["Toggle Mount"]
BINDING_NAME_POKEDEXSUMMONOTHERMOUNT = L["Summon Other Mount"]
BINDING_NAME_POKEDEXSUMMONCOMPANION = L["Summon Companion"]
BINDING_NAME_POKEDEXDISMISSCOMPANION = L["Dismiss Companion"]
BINDING_NAME_POKEDEXTOGGLECOMPANION = L["Toggle Companion"]

-- Announcement Info --
local g_rgChannelDescs = { L["personal"], L["party"], L["raid"], L["emote"], L["say"], L["yell"] }
local g_rgChannels = { "ERROR", "PARTY", "RAID", "EMOTE", "SAY", "YELL" } -- CHATID types, not localized, ERROR is just placeholder

-- Title Info -- 
local TI_NAME = 1
local TI_ID = 2
local TI_RANK = 3

-- Companion Info -- 
local PI_NAME = 1
local PI_ID = 2
local PI_RANK = 3
local PI_TYPE = 4

local MT_FF = 1
local MT_SF = 2
local MT_FR = 3
local MT_SR = 4

local g_iCurrentMount = 0
local g_iMountForRanking = 0
local g_iHotMount = 0
local g_rgstrMountNames = {}
local g_rgstrHotMountNames = {}

local g_iCurrentCompanion = 0
local g_iCompanionForRanking = 0
local g_iHotCompanion = 0
local g_rgstrCompanionNames = {}
local g_rgstrHotCompanionNames = {}

local g_iTitleForRanking = 0
local g_iHotTitle = 0
local g_rgstrTitleNames = {}
local g_rgstrHotTitleNames = {}

local g_iDefaultRank = 5
local g_iCurDataFormat = 1

local defaults = {
	profile = {
		iDataFormat = 0,
		fFirstBoot = true,
		fAnnounce = true,
		iChannel = 1,
		fChangeTitleOnLoad = true,
		fEnableHotness = true,
		fNewHotness = true,
		idHotMount = 0,
		iHotMountPref = 50,
		idHotCompanion = 0,
		iHotCompanionPref = 50,
		idHotTitle = 0,
		iHotTitlePref = 50,
		rgMountInfo = {},
		rgCompanionInfo = {},
		rgTitleInfo = {}
	},
}

-- As the addon develops and ages, there may be changes to the structure 
-- and contents of the saved data. This function migrates data between
-- data format versions to preserve the users settings.
function Pokedex:MigrateSavedData()
	-- if its first boot, theres no data to migrate
	if not self.db.profile.fFirstBoot then
		
		if self.db.profile.iDataFormat < 1 then
			-- rename of critter to companion
			if self.db.profile.idHotCritter ~= nil then
				self.db.profile.idHotCompanion = self.db.profile.idHotCritter
				self.db.profile.idHotCritter = nil
			end
			if self.db.profile.idHotCritterPref ~= nil then
				self.db.profile.iHotCompanionPref = self.db.profile.idHotCritterPref
				self.db.profile.idHotCritterPref = nil
			end
			if self.db.profile.rgCritterInfo ~= nil then
				self.db.profile.rgCompanionInfo = self.db.profile.rgCritterInfo
				self.db.profile.rgCritterInfo = nil
			end
		end
	end
			
	self.db.profile.iDataFormat = g_iCurDataFormat
end


local options = {
    name = "Pokedex",
    handler = Pokedex,
    type = 'group',
    childGroups = "tab",
    args = {
		zone = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["echoes out zone info"],
			desc = L["echoes out zone info"],
			func = "EchoZone",
		},
		SuperToggle = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["toggle mount or companion"],
			desc = L["toggles mount or companion"],
			func = "SuperToggle",
		},
		SummonMount = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["summon mount"],
			desc = L["summons a mount"],
			func = "SummonMount",
		},
		sm = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["summon mount"],
			desc = L["summons a mount"],
			func = "SummonMount",
		},
		DismissMount = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["dismiss mount"],
			desc = L["dismisses current mount"],
			func = "DismissMount",
		},
		dm = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["dismiss mount"],
			desc = L["dismisses current mount"],
			func = "DismissMount",
		},
		ToggleMount = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["toggle mount"],
			desc = L["toggles a mount"],
			func = "ToggleMount",
		},
		tm = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["toggle mount"],
			desc = L["toggles a mount"],
			func = "ToggleMount",
		},
		SummonOtherMount = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["summon other mount"],
			desc = L["if summon mount would summon flyer, will summon ground mount and vice versa"],
			func = "SummonOtherMount",
		},
		som = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["summon other mount"],
			desc = L["if summon mount would summon flyer, will summon ground mount and vice versa"],
			func = "SummonOtherMount",
		},
		SummonCompanion = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["summon companion"],
			desc = L["summons a companion"],
			func = "SummonCompanion",
		},
		sc = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["summon companion"],
			desc = L["summons a companion"],
			func = "SummonCompanion",
		},
		DismissCompanion = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["dismiss companion"],
			desc = L["dismisses current companion"],
			func = "DismissCompanion",
		},
		dc = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["dismiss companion"],
			desc = L["dismisses current companion"],
			func = "DismissCompanion",
		},
		ToggleCompanion = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["toggle companion"],
			desc = L["toggles a companion"],
			func = "ToggleCompanion",
		},
		tc = {
			hidden = true,
			order = 0,
			type = "execute",
			name = L["toggle companion"],
			desc = L["toggles a companion"],
			func = "ToggleCompanion",
		},

		CtrlAnnounce = {
			order = 1,
			type = "toggle",
			name = L["announce"],
			desc = L["(on|off) let everyone know who *you* choose"],
			get = "FGetAnnounce",
			set = "SetAnnounce",
		},
		CtrlChannel = {
			order = 2,
			type = "select",
			name = L["channel"],
			desc = L["channel to announce selection in"],
			values = g_rgChannelDescs,
			get = "GetChannel",
			set = "SetChannel",
		},
		CtrlEnableHotness = {
			order = 3,
			type = "toggle",
			name = L["enable hot pets"],
			desc = L["lets your turn on|off the hot pet subfeatures"],
			get = "FGetEnableHotness",
			set = "SetEnableHotness",
		},
		CtrlNewHotness = {
			order = 4,
			type = "toggle",
			name = L["new hotness"],
			desc = L["always make newest mount or companion the hot one"],
			disabled = "FGetDisableHotness",
			get = "FGetNewHotness",
			set = "SetNewHotness",
		},

		CtrlGroupMounts = {
			order = 20,
			name = "Mounts",
			handler = Pokedex,
			type = 'group',
			args = {
				CtrlSummonMount = {
					order = 1,
					type = "execute",
					name = L["summon mount"],
					desc = L["summons a mount"],
					func = "SummonMount",
				},
				CtrlDismissMount = {
					order = 2,
					type = "execute",
					name = L["dismiss mount"],
					desc = L["dismisses current mount"],
					func = "DismissMount",
				},
				CtrlMountForRanking = {
					order = 3,
					type = "select",
					name = L["select mount for ranking"],
					desc = L["mount whose rank you can set"],
					values = g_rgstrMountNames,
					get = "GetMountForRanking",
					set = "SetMountForRanking",
				},
				CtrlRankingForMount = {
					order = 4,
					type = "range",
					name = L["mount rank"],
					desc = L["rank of current mount"],
					min = 0,
					max = 10,
					step = 1,
					disabled = "FDisablePerMountSettings",
					get = "GetRankingForMount",
					set = "SetRankingForMount",
				},
				CtrlFastFlyer = {
					order = 7,
					type = "toggle",
					name = L["Fast Flyer"],
					desc = L["mount has a fast (280%) flight speed"],
					disabled = "FDisablePerMountSettings",
					get = "FGetFastFlyer",
					set = "SetFastFlyer",
				},
				CtrlSlowFlyer = {
					order = 8,
					type = "toggle",
					name = L["Slow Flyer"],
					desc = L["mount has a slow (60%) flight speed"],
					disabled = "FDisablePerMountSettings",
					get = "FGetSlowFlyer",
					set = "SetSlowFlyer",
				},
				CtrlFastRunner = {
					order = 9,
					type = "toggle",
					name = L["Fast Runner"],
					desc = L["mount has a fast (100%) running speed *AND* can be summoned in no-fly zones"],
					disabled = "FDisablePerMountSettings",
					get = "FGetFastRunner",
					set = "SetFastRunner",
				},
				CtrlSlowRunner = {
					order = 10,
					type = "toggle",
					name = L["Slow Runner"],
					desc = L["mount has a slow (60%) running speed *AND* can be summoned in no-fly zones"],
					disabled = "FDisablePerMountSettings",
					get = "FGetSlowRunner",
					set = "SetSlowRunner",
				},
				CtrlHotMount = {
					order = 11,
					type = "select",
					name = L["select hot mount"],
					desc = L["mount to become hot one"],
					values = g_rgstrHotMountNames,
					disabled = "FGetDisableHotness",
					get = "GetHotMount",
					set = "SetHotMount",
				},
				CtrlMountHeat = {
					order = 12,
					type = "range",
					name = L["mount heat"],
					desc = L["set hotness as a percentage - 100% means the hot pet is the only one that will be chosen"],
					min = 0,
					max = 100,
					step = 5,
					disabled = "FGetDisableHotness",
					get = "GetMountHeat",
					set = "SetMountHeat",
				},
			},
		},
		CtrlGroupCompanions = {
			order = 40,
			name = "Companions",
			handler = Pokedex,
			type = 'group',
			args = {
				CtrlSummonCompanion = {
					order = 2,
					type = "execute",
					name = L["summon companion"],
					desc = L["summons a companion"],
					func = "SummonCompanion",
				},
				CtrlDismissCompanion = {
					order = 3,
					type = "execute",
					name = L["dismiss companion"],
					desc = L["dismisses current companion"],
					func = "DismissCompanion",
				},
				CtrlCompanionForRanking = {
					order = 4,
					type = "select",
					name = L["select companion for ranking"],
					desc = L["companion whose rank you can set"],
					values = g_rgstrCompanionNames,
					get = "GetCompanionForRanking",
					set = "SetCompanionForRanking",
				},
				CtrlRankingForCompanion = {
					order = 5,
					type = "range",
					name = L["companion rank"],
					desc = L["rank of current companion"],
					min = 0,
					max = 10,
					step = 1,
					disabled = "FDisableRankingForCompanion",
					get = "GetRankingForCompanion",
					set = "SetRankingForCompanion",
				},
				CtrlHotCompanion = {
					order = 6,
					type = "select",
					name = L["select hot companion"],
					desc = L["companion to become hot one"],
					values = g_rgstrHotCompanionNames,
					disabled = "FGetDisableHotness",
					get = "GetHotCompanion",
					set = "SetHotCompanion",
				},
				CtrlCompanionHeat = {
					order = 7,
					type = "range",
					name = L["companion heat"],
					desc = L["set hotness as a percentage - 100% means the hot pet is the only one that will be chosen"],
					min = 0,
					max = 100,
					step = 5,
					disabled = "FGetDisableHotness",
					get = "GetCompanionHeat",
					set = "SetCompanionHeat",
				},
			},
		},
		CtrlGroupTitles = {
			order = 60,
			name = "Titles",
			handler = Pokedex,
			type = 'group',
			args = {
				CtrlChangeTitleOnLoad = {
					order = 2,
					type = "toggle",
					name = L["Change title on load"],
					desc = L["Change title when logging in or reloading UI"],
					get = "FGetChangeTitleOnLoad",
					set =  "SetChangeTitleOnLoad",
				},
				CtrlChangeTitle = {
					order = 3,
					type = "execute",
					name = L["change title"],
					desc = L["change title"],
					func = "ChangeTitle",
				},
				CtrlTitleForRanking = {
					order = 4,
					type = "select",
					name = L["select title for ranking"],
					desc = L["title whose rank you can set"],
					values = g_rgstrTitleNames,
					get = "GetTitleForRanking",
					set = "SetTitleForRanking",
				},
				CtrlRankingForTitle = {
					order = 5,
					type = "range",
					name = L["title rank"],
					desc = L["rank of current title"],
					min = 0,
					max = 10,
					step = 1,
					disabled = "FDisableRankingForTitle",
					get = "GetRankingForTitle",
					set = "SetRankingForTitle",
				},
				CtrlHotTitle = {
					order = 6,
					type = "select",
					name = L["select hot title"],
					desc = L["title to become hot one"],
					values = g_rgstrHotTitleNames,
					disabled = "FGetDisableHotness",
					get = "GetHotTitle",
					set = "SetHotTitle",
				},
				CtrlTitleHeat = {
					order = 7,
					type = "range",
					name = L["title heat"],
					desc = L["set hotness as a percentage - 100% means the hot pet is the only one that will be chosen"],
					min = 0,
					max = 100,
					step = 5,
					disabled = "FGetDisableHotness",
					get = "GetTitleHeat",
					set = "SetTitleHeat",
				},
			},
		},
    },
}


-- Called when the addon is loaded
function Pokedex:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("PokedexDB", defaults)

	--self.optionsFrame = LibStub("AceConfig-3.0"):AddToBlizOptions("Pokedex", "Pokedex")
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Pokedex", options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Pokedex", "Pokedex")
	self:RegisterChatCommand("pd", "ChatCommand")
	self:RegisterChatCommand("pokedex", "ChatCommand")
end


function Pokedex:ChatCommand(input)
	if not input or input:trim() == "" then
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
		--LibStub("AceConfigDialog-3.0"):Open("Pokedex")
    else
        LibStub("AceConfigCmd-3.0").HandleCommand(Pokedex, "pd", "Pokedex", input)
    end
end


-- Called when the addon is enabled
function Pokedex:OnEnable()
	CreateFrame( "GameTooltip", "PokedexScanningTooltip" ); -- Tooltip name cannot be nil
	PokedexScanningTooltip:SetOwner( WorldFrame, "ANCHOR_NONE" );
	-- Allow tooltip SetX() methods to dynamically add new lines based on these
	PokedexScanningTooltip:AddFontStrings(
		PokedexScanningTooltip:CreateFontString( "$parentTextLeft1", nil, "GameTooltipText" ),
		PokedexScanningTooltip:CreateFontString( "$parentTextRight1", nil, "GameTooltipText" ) );

	self:MigrateSavedData()

	self:UpdateMountInfo()
	self:UpdateCompanionInfo()
	self:UpdateTitleInfo()

	self.db.profile.fFirstBoot = false
	self:EchoPetCount()
	
	-- change title on load
	if (self.db.profile.fChangeTitleOnLoad) then self:ChangeTitle() end

	-- update data when new pet added	
	self:RegisterEvent("COMPANION_LEARNED")

	-- keep track of currently visible
	self:RegisterEvent("COMPANION_UPDATE")

	-- update data when new data added	
	self:RegisterEvent("NEW_TITLE_EARNED")

end

-- Called when the addon is disabled
function Pokedex:OnDisable()
end


function Pokedex:COMPANION_LEARNED()
	self:UpdateMountInfo()
	self:UpdateCompanionInfo()
	-- self:EchoCompanionCount()
end


function Pokedex:COMPANION_UPDATE()
	self:GetCurrentMount()
	self:GetCurrentCompanion()
end


function Pokedex:NEW_TITLE_EARNED()
	self:UpdateTitleInfo()
end


function Pokedex:SuperToggle() 
	if IsOutdoors() then
		self:ToggleMount()
	else
		self:ToggleCompanion()
	end
end

function Pokedex:EchoPetCount()
	self:Print("Current pet counts are " .. #self.db.profile.rgMountInfo .. " mounts and " .. #self.db.profile.rgCompanionInfo .. " companions")
end

function Pokedex:EchoZone()
	SetMapToCurrentZone()
	self:Print(GetCurrentMapContinent() .. "  " .. GetZoneText() .. "  " .. GetSubZoneText())
end

--=========================================================================--
-- Announcement functions
--=========================================================================--

function Pokedex:FGetAnnounce(info) return self.db.profile.fAnnounce; end
function Pokedex:SetAnnounce(info, fNew) self.db.profile.fAnnounce = fNew; end

function Pokedex:GetChannel(info) return self.db.profile.iChannel; end
function Pokedex:SetChannel(info, iNew)
	if (iNew >= 1 and iNew <= 6) then
		self.db.profile.iChannel = iNew
	else
		self:Print(L["ERROR: bad channel set"])
	end
end

function Pokedex:AnnounceSummon(iNew)
	if (not self.db.profile.fAnnounce) then return end

	local iChannel = self.db.profile.iChannel
	
	-- if not in raid, try sending it to party
	if (3 == iChannel and 0 == GetNumRaidMembers()) then
		iChannel = 2
	end
	-- if not in a party, just send it to yourself
	if (2 == iChannel and 0 == GetNumPartyMembers()) then
		iChannel = 1
	end
	
	if (iChannel == 1) then
		self:Print(L['X, I choose you!'](g_rgstrCompanionNames[iNew]));
	elseif (iChannel == 4) then
		SendChatMessage(L['lets X know that they have been chosen.'](g_rgstrCompanionNames[iNew]), "EMOTE");
	else
		SendChatMessage(L['X, I choose you!'](g_rgstrCompanionNames[iNew]), g_rgChannels[iChannel]);
	end
	
end

--=========================================================================--
-- general Hotness functions
--=========================================================================--

function Pokedex:FGetEnableHotness(info) return self.db.profile.fEnableHotness; end
function Pokedex:SetEnableHotness(info, fNew) self.db.profile.fEnableHotness = fNew; end
function Pokedex:FGetDisableHotness(info) return not self:FGetEnableHotness(info); end
function Pokedex:FGetNewHotness(info) return self.db.profile.fNewHotness; end
function Pokedex:SetNewHotness(info, fNew) self.db.profile.fNewHotness = fNew; end

--=========================================================================--
-- Mount functions
--=========================================================================--

function Pokedex:SummonMount() self:SummonMountInner(false); end
function Pokedex:SummonOtherMount() self:SummonMountInner(true); end

function Pokedex:SummonMountInner(fOther)
	local iCurrent = self:GetCurrentMount()
	local iMountType
	local fFlyable = false
	
	if (iCurrent ~= 0) then
		local start, duration, enable = GetCompanionCooldown("MOUNT", iCurrent)
		if (duration ~= 0) then
			-- self:Print("in cooldown, ignore this summons attempt")
			return
		end
	end

	local rgcMountRanks = { 0, 0, 0, 0}
	for iRanks = 1, #self.db.profile.rgMountInfo, 1 do
		for iType = MT_FF, #rgcMountRanks, 1 do
			if (self.db.profile.rgMountInfo[iRanks][PI_TYPE][iType]) then
				rgcMountRanks[iType] = rgcMountRanks[iType] + self.db.profile.rgMountInfo[iRanks][PI_RANK]
			end
		end
	end

	-- check location, what kind of mounts can we actually summon?
	SetMapToCurrentZone()
	continent = GetCurrentMapContinent()
	
	if 3 == continent then fFlyable = true end
	if 4 == continent and IsUsableSpell(L["Cold Weather Flying"]) then
		local strZone = GetZoneText()
		if strfind(strZone, L["Dalaran"]) then
			fFlyable = strfind(GetSubZoneText(), L["Krasus' Landing"])
		else
			fFlyable = not strfind(strZone, L["Wintergrasp"]) and not strfind(strZone, L["The Obsidian Sanctum"])
		end
	end

	-- flip the type if the user has asked for that
	if fOther then fFlyable = not fFlyable end
	
	if fFlyable then -- flyers
		for iType = MT_FF, #rgcMountRanks, 1 do
			if (rgcMountRanks[iType] > 0) then
				iMountType = iType
				break
			end
		end
	else -- runners
		for iType = MT_FR, #rgcMountRanks, 1 do
			if (rgcMountRanks[iType] > 0) then
				iMountType = iType
				break
			end
		end
	end

	local cRanks = rgcMountRanks[iMountType]

	-- hot pet check, only matters if hot pet is in set we're summoning from
	if (self.db.profile.fEnableHotness and g_iHotMount ~= 0 and g_iHotMount ~= iCurrent and self.db.profile.rgMountInfo[g_iHotMount][PI_TYPE][iMountType]) then
		-- do hot pet check and possibly summon
		if (self.db.profile.iHotMountPref >= math.random(1,100)) then
			CallCompanion("Mount", g_iHotMount)
			-- TODO self:AnnounceSummon(g_iHotMount)
			return
		end
		
		-- if you're here, they have a hot pet, but its not coming out
		cRanks = cRanks - self.db.profile.rgMountInfo[g_iHotMount][PI_RANK]
	end
	
	-- if the current mount is in the set we're summoning out of, remove it from the ranks total
	if (iCurrent ~= 0 and self.db.profile.rgMountInfo[iCurrent][PI_TYPE][iMountType]) then
		cRanks = cRanks - self.db.profile.rgMountInfo[iCurrent][PI_RANK]
	end

	if (cRanks <= 0) then 
		if (iCurrent ~= 0) then
			return -- only summonable pet is already summoned
		end
		
		self:Print(L["ERROR: You have no summonable mounts."])
		return
	end
	
	cRanks = math.random(1,cRanks)

	for iNew = 1, #self.db.profile.rgMountInfo, 1 do
		if (self.db.profile.rgMountInfo[iNew][PI_TYPE][iMountType] 
		 and iNew ~= iCurrent and (iNew ~= g_iHotMount or not self.db.profile.fEnableHotness)) then
			cRanks = cRanks - self.db.profile.rgMountInfo[iNew][PI_RANK]
			if (cRanks <= 0) then 
				-- found our slot
				CallCompanion("MOUNT", iNew)
				-- TODO self:AnnounceSummon(iNew)
				return
			end  
		end
	end
	
	if (cRanks > 0) then self:Print(L["ERROR: selection error"]) end
end


function Pokedex:DismissMount()
	if UnitInVehicle("player") then
		VehicleExit()
	else
		Dismount()
	end
end


function Pokedex:ToggleMount()
	if (self:GetCurrentMount() > 0 or UnitInVehicle("player")) then
		self:DismissMount()
	else
		self:SummonMount()
	end
end


function Pokedex:GetCurrentMount()
	local creatureID, creatureName, spellID, icon, active

	if (g_iCurrentMount ~= nil and g_iCurrentMount~=0) then
		creatureID, creatureName, spellID, icon, active = GetCompanionInfo("MOUNT", g_iCurrentMount)
		if (active) then 
			return g_iCurrentMount
		end 
	end
	
	g_iCurrentMount = 0 -- might not be any Mount active anymore
	
	for index = 1, GetNumCompanions("MOUNT"), 1 do
		creatureID, creatureName, spellID, icon, active = GetCompanionInfo("MOUNT", index)
		if active then 
			g_iCurrentMount = index
			break
		end
	end
	
	return g_iCurrentMount
end


function Pokedex:UpdateMountInfo()
	g_iCurrentMount = 0
	g_iHotMount = 0
	
	-- assumption is that order doesn't change except for insertions of new pets
	-- so we can speed up searching saved data for rankings by starting the search 
	-- at a predicted index
	local iMatch = 1
	local rgNew = {}
	
	for index = 1, GetNumCompanions("MOUNT"), 1 do
		local creatureID, creatureName, spellID, icon, active = GetCompanionInfo("MOUNT", index)
		local fMatchMade, fFF, fSF, fFR, fSR, iRank = self:GetMountRankAndTypesFromSavedData(iMatch, creatureID)

		if (fMatchMade) then 
			iMatch = iMatch + 1 
		else
			-- make a guess as to type info
			fFF, fSF, fFR, fSR = self:MountTypesFromName(creatureName, spellID)
			
			if (not self.db.profile.fFirstBoot) then
				self:Print(L["New mount added! "] .. creatureName)
				if (self.db.profile.fEnableHotness and self.db.profile.fNewHotness) then
					self.db.profile.idHotMount = creatureID
				end
			end
		end
		
		
		if (creatureID == self.db.profile.idHotMount) then g_iHotMount = index end
		if (active) then g_iCurrentMount = index end
		
		g_rgstrMountNames[index] = creatureName
		g_rgstrHotMountNames[index] = creatureName
		rgNew[index] = { creatureName, creatureID, iRank, { fFF, fSF, fFR, fSR } }
	end

	if (#g_rgstrMountNames < 1) then
		g_rgstrMountNames[1] = L["no mounts available"]
	end

	g_rgstrHotMountNames[0] = L["no hot mount"]

	
	-- now that we've read everything in, blow away old saved ranking with new list
	self.db.profile.rgMountInfo = rgNew
	
	if (g_iCurrentMount > 0) then g_iMountForRanking = g_iCurrentMount else g_iMountForRanking = 1 end
end


-- return values - fMatchMade, fFastFlyer, fSlowFlyer, fFastRunner, fSlowRunner, iRank 
function Pokedex:GetMountRankAndTypesFromSavedData(iStart, iID)
	local cLimit = #self.db.profile.rgMountInfo
	local index
	
	if (iStart <= cLimit) then
		for index = iStart, cLimit, 1 do
			if (self.db.profile.rgMountInfo[index][PI_ID] == iID) then
				return true, self.db.profile.rgMountInfo[index][PI_TYPE][MT_FF], self.db.profile.rgMountInfo[index][PI_TYPE][MT_SF], self.db.profile.rgMountInfo[index][PI_TYPE][MT_FR], self.db.profile.rgMountInfo[index][PI_TYPE][MT_SR], self.db.profile.rgMountInfo[index][PI_RANK]
			end
		end
		
		for index = 1, iStart - 1, 1 do
			if (self.db.profile.rgMountInfo[index][PI_ID] == iID) then
				return true, self.db.profile.rgMountInfo[index][PI_TYPE][MT_FF], self.db.profile.rgMountInfo[index][PI_TYPE][MT_SF], self.db.profile.rgMountInfo[index][PI_TYPE][MT_FR], self.db.profile.rgMountInfo[index][PI_TYPE][MT_SR], self.db.profile.rgMountInfo[index][PI_RANK]
			end
		end
	end
	
	return false, false, false, false, false, g_iDefaultRank
end


function Pokedex:MountTypesFromName(creatureName, spellID)
	-- TODO handle speeds like extremely fast, slow and barely moving
	-- TODO handle rider skill based mounts
	
	-- handle some of the ones that don't follow normal patterns seperately by name
	if (strfind(creatureName, "Ashes of Al'ar") or
		strfind(creatureName, "Black Proto-Drake") or
		strfind(creatureName, "Plagued Proto-Drake") or
		strfind(creatureName, "Violet Proto-Drake")) then
		return true, true, false, false

	elseif (strfind(creatureName, "Winged Steed of the Ebon Blade")) then
		return true, true, false, false

	elseif (strfind(creatureName, "Headless Horseman's Mount")) then
		return true, true, true, true
	elseif (strfind(creatureName, "Big Blizzard Bear")) then
		return false, false, true, true

	elseif (strfind(creatureName, "Black Qiraji Battle Tank")) then
		return false, false, true, true
	elseif (strfind(creatureName, "Qiraji")) then
		return false, false, false, false
	end
	
	local fast = false
	local flyer = false
	
	PokedexScanningTooltip:ClearLines()
	PokedexScanningTooltip:SetHyperlink("spell:"..spellID) 
	
	for iLines=1,PokedexScanningTooltip:NumLines() do
		local TipTextObject = getglobal("PokedexScanningTooltipTextLeft" .. iLines)
		local strText = TipTextObject:GetText()
		self:Print(strText)
		
		if strfind(strText, L["very fast"]) then 
			fast = true 
		elseif strfind(strText, L["extremely fast"]) then 
			fast = true
		end
		
		if strfind(strText, L["only be summoned in Outland or Northrend"]) then 
			flyer = true
		elseif strfind(strText, L["only be used in Outland and Northrend"]) then 
			flyer = true
		end
	end
	
	return (flyer and fast), flyer, (fast and not flyer), not flyer
end

--=========================================================================--
-- Mount options UI functions
--=========================================================================--

function Pokedex:GetMountForRanking(info) return g_iMountForRanking; end
function Pokedex:SetMountForRanking(info, iNew) g_iMountForRanking = iNew; end


function Pokedex:GetRankingForMount(info)
	if (#self.db.profile.rgMountInfo > 0) then
		return self.db.profile.rgMountInfo[g_iMountForRanking][PI_RANK]
	else
		return g_iDefaultRank
	end
end

function Pokedex:SetRankingForMount(info, iNew)
	self.db.profile.rgCompanionInfo[g_iCompanionForRanking][PI_RANK] = iNew
end

function Pokedex:FDisablePerMountSettings(info) return not (#self.db.profile.rgMountInfo > 0); end

function Pokedex:FGetFastFlyer(info) 
	if (#self.db.profile.rgMountInfo > 0) then
		return self.db.profile.rgMountInfo[g_iMountForRanking][PI_TYPE][MT_FF]
	end
	return false
end

function Pokedex:FGetSlowFlyer(info) 
	if (#self.db.profile.rgMountInfo > 0) then
		return self.db.profile.rgMountInfo[g_iMountForRanking][PI_TYPE][MT_SF]
	end
	return false
end

function Pokedex:FGetFastRunner(info) 
	if (#self.db.profile.rgMountInfo > 0) then
		return self.db.profile.rgMountInfo[g_iMountForRanking][PI_TYPE][MT_FR]
	end
	return false
end

function Pokedex:FGetSlowRunner(info) 	if (#self.db.profile.rgMountInfo > 0) then
		return self.db.profile.rgMountInfo[g_iMountForRanking][PI_TYPE][MT_SR]
	end
	return false
end

function Pokedex:SetFastFlyer(info, fNew) self.db.profile.rgMountInfo[g_iMountForRanking][PI_TYPE][MT_FF] = fNew; end
function Pokedex:SetSlowFlyer(info, fNew) self.db.profile.rgMountInfo[g_iMountForRanking][PI_TYPE][MT_SF] = fNew; end
function Pokedex:SetFastRunner(info, fNew) self.db.profile.rgMountInfo[g_iMountForRanking][PI_TYPE][MT_FR] = fNew; end
function Pokedex:SetSlowRunner(info, fNew) self.db.profile.rgMountInfo[g_iMountForRanking][PI_TYPE][MT_SR] = fNew; end

function Pokedex:GetHotMount(info) return g_iHotMount; end
function Pokedex:SetHotMount(info, iNew)
	g_iHotMount = iNew
	if (iNew > 0) then
		self.db.profile.idHotMount = self.db.profile.rgMountInfo[iNew][PI_ID]
	else
		self.db.profile.idHotMount = 0;
	end
end

function Pokedex:GetMountHeat(info) return self.db.profile.iHotMountPref; end
function Pokedex:SetMountHeat(info, iNew) self.db.profile.iHotMountPref = iNew; end


--=========================================================================--
-- Companion functions
--=========================================================================--

function Pokedex:SummonCompanion()
	local iCurrent = self:GetCurrentCompanion()
	
	if (iCurrent ~= 0) then
		local start, duration, enable = GetCompanionCooldown("CRITTER", iCurrent)
		if (duration ~= 0) then
			-- self:Print("in cooldown, ignore this summons attempt")
			return
		end
	end

	local cRanks = 0
	for iRanks = 1, #self.db.profile.rgCompanionInfo, 1 do
		cRanks = cRanks + self.db.profile.rgCompanionInfo[iRanks][PI_RANK]
	end

	if (self.db.profile.fEnableHotness and g_iHotCompanion ~= 0 and g_iHotCompanion ~= iCurrent) then
		-- do hot pet check and possibly summon
		if (self.db.profile.iHotCompanionPref >= math.random(1,100)) then
			CallCompanion("CRITTER", g_iHotCompanion)
			self:AnnounceSummon(g_iHotCompanion)
			return
		end
		
		-- if you're here, they have a hot pet, but its not coming out
		cRanks = cRanks - self.db.profile.rgCompanionInfo[g_iHotCompanion][PI_RANK]
	end
	
	if (iCurrent ~= 0) then
		cRanks = cRanks - self.db.profile.rgCompanionInfo[iCurrent][PI_RANK]
	end

	if (cRanks <= 0) then 
		if (iCurrent ~= 0) then
			return -- only summonable pet is already summoned
		end
		
		self:Print(L["ERROR: You have no summonable companions."])
		return
	end
	
	cRanks = math.random(1,cRanks)

	for iNew = 1, #self.db.profile.rgCompanionInfo, 1 do
		if (iNew ~= iCurrent and (iNew ~= g_iHotCompanion or not self.db.profile.fEnableHotness)) then
			cRanks = cRanks - self.db.profile.rgCompanionInfo[iNew][PI_RANK]
			if (cRanks <= 0) then 
				-- found our slot
				CallCompanion("CRITTER", iNew)
				self:AnnounceSummon(iNew)
				return
			end  
		end
	end
	
	if (cRanks > 0) then self:Print(L["ERROR: selection error"]) end
end


function Pokedex:DismissCompanion()
	local iCurrent = self:GetCurrentCompanion()
	if (iCurrent > 0) then
		CallCompanion("CRITTER", iCurrent) 
	end 
end


function Pokedex:ToggleCompanion()
	if (self:GetCurrentCompanion() > 0) then
		self:DismissCompanion()
	else
		self:SummonCompanion()
	end
end


function Pokedex:GetCurrentCompanion()
	local creatureID, creatureName, spellID, icon, active

	if (g_iCurrentCompanion ~= nil and g_iCurrentCompanion~=0) then
		creatureID, creatureName, spellID, icon, active = GetCompanionInfo("CRITTER", g_iCurrentCompanion)
		if (active) then 
			return g_iCurrentCompanion
		end 
	end
	
	g_iCurrentCompanion = 0 -- might not be any companion active anymore
	
	for index = 1, GetNumCompanions("CRITTER"), 1 do
		creatureID, creatureName, spellID, icon, active = GetCompanionInfo("CRITTER", index)
		if active then 
			g_iCurrentCompanion = index
			break
		end
	end
	
	return g_iCurrentCompanion
end


function Pokedex:UpdateCompanionInfo()
	g_iCurrentCompanion = 0
	g_iHotCompanion = 0
	
	-- assumption is that order doesn't change except for insertions of new pets
	-- so we can speed up searching saved data for rankings by starting the search 
	-- at a predicted index
	local iMatch = 1
	local rgNew = {}
	
	for index = 1, GetNumCompanions("CRITTER"), 1 do
		local creatureID, creatureName, spellID, icon, active = GetCompanionInfo("CRITTER", index)
		local fMatchMade, iRank = self:GetCompanionRankFromSavedData(iMatch, creatureID)

		if (fMatchMade) then 
			iMatch = iMatch + 1 
		elseif (not self.db.profile.fFirstBoot) then
			self:Print(L["New companion added! "] .. creatureName)
			if (self.db.profile.fEnableHotness and self.db.profile.fNewHotness) then
				self.db.profile.idHotCompanion = creatureID
			end
		end
		
		
		if (creatureID == self.db.profile.idHotCompanion) then g_iHotCompanion = index end
		if (active) then g_iCurrentCompanion = index end
		
		g_rgstrCompanionNames[index] = creatureName
		g_rgstrHotCompanionNames[index] = creatureName
		rgNew[index] = { creatureName, creatureID, iRank }
	end

	if (#g_rgstrCompanionNames < 1) then
		g_rgstrCompanionNames[1] = L["no companions available"]
	end

	g_rgstrHotCompanionNames[0] = L["no hot companion"]

	
	-- now that we've read everything in, blow away old saved ranking with new list
	self.db.profile.rgCompanionInfo = rgNew
	
	if (g_iCurrentCompanion > 0) then g_iCompanionForRanking = g_iCurrentCompanion else g_iCompanionForRanking = 1 end
end


function Pokedex:GetCompanionRankFromSavedData(iStart, iID)
	local cLimit = #self.db.profile.rgCompanionInfo
	local index
	
	if (iStart <= cLimit) then
		for index = iStart, cLimit, 1 do
			if (self.db.profile.rgCompanionInfo[index][PI_ID] == iID) then
				return true, self.db.profile.rgCompanionInfo[index][PI_RANK]
			end
		end
		
		for index = 1, iStart - 1, 1 do
			if (self.db.profile.rgCompanionInfo[index][PI_ID] == iID) then
				return true, self.db.profile.rgCompanionInfo[index][PI_RANK]
			end
		end
	end
	
	return false, g_iDefaultRank
end


--=========================================================================--
-- Companion options UI functions
--=========================================================================--

function Pokedex:GetCompanionForRanking(info) return g_iCompanionForRanking; end
function Pokedex:SetCompanionForRanking(info, iNew) g_iCompanionForRanking = iNew; end


function Pokedex:GetRankingForCompanion(info)
	if (#self.db.profile.rgCompanionInfo > 0) then
		return self.db.profile.rgCompanionInfo[g_iCompanionForRanking][PI_RANK]
	else
		return g_iDefaultRank
	end
end

function Pokedex:SetRankingForCompanion(info, iNew)
	self.db.profile.rgCompanionInfo[g_iCompanionForRanking][PI_RANK] = iNew
end


function Pokedex:FDisableRankingForCompanion(info) return not (#self.db.profile.rgCompanionInfo > 0); end


function Pokedex:GetHotCompanion(info) return g_iHotCompanion; end
function Pokedex:SetHotCompanion(info, iNew)
	g_iHotCompanion = iNew
	if (iNew > 0) then
		self.db.profile.idHotCompanion = self.db.profile.rgCompanionInfo[iNew][PI_ID]
	else
		self.db.profile.idHotCompanion = 0;
	end
end

function Pokedex:GetCompanionHeat(info) return self.db.profile.iHotCompanionPref; end
function Pokedex:SetCompanionHeat(info, iNew) self.db.profile.iHotCompanionPref = iNew; end


--=========================================================================--
-- Title functions
--=========================================================================--

function Pokedex:ChangeTitle()
	local iCurrentTitle = GetCurrentTitle()
	local iCurrent = 0
	
	local cRanks = 0
	for iRanks = 1, #self.db.profile.rgTitleInfo, 1 do
		cRanks = cRanks + self.db.profile.rgTitleInfo[iRanks][TI_RANK]
		if self.db.profile.rgTitleInfo[iRanks][TI_ID] == iCurrentTitle then
			iCurrent = iRanks
		end
	end

	if (self.db.profile.fEnableHotness and g_iHotTitle ~= 0 and g_iHotTitle ~= iCurrent) then
		-- do hot check and possibly change
		if (self.db.profile.iHotTitlePref >= math.random(1,100)) then
			SetCurrentTitle(self.db.profile.rgTitleInfo[g_iHotTitle][TI_ID])
			return
		end
		
		-- if you're here, they have a hot title, but its getting set
		cRanks = cRanks - self.db.profile.rgTitleInfo[g_iHotTitle][TI_RANK]
	end
	
	if (iCurrent ~= 0) then
		cRanks = cRanks - self.db.profile.rgTitleInfo[iCurrent][TI_RANK]
	end

	if (cRanks <= 0) then 
		if (iCurrent ~= 0) then
			return -- only title is already displayed
		end
		
		-- self:Print(L["ERROR: You don't have any titles."])
		return
	end
	
	cRanks = math.random(1,cRanks)

	for iNew = 1, #self.db.profile.rgTitleInfo, 1 do
		if (iNew ~= iCurrent and (iNew ~= g_iHotTitle or not self.db.profile.fEnableHotness)) then
			cRanks = cRanks - self.db.profile.rgTitleInfo[iNew][TI_RANK]
			if (cRanks <= 0) then 
				-- found our slot
				SetCurrentTitle(self.db.profile.rgTitleInfo[iNew][TI_ID])
				return
			end  
		end
	end
	
	if (cRanks > 0) then self:Print(L["ERROR: selection error"]) end
end


function Pokedex:UpdateTitleInfo()
	g_iHotTitle = 0
	
	-- assumption is that order doesn't change except for insertions of new pets
	-- so we can speed up searching saved data for rankings by starting the search 
	-- at a predicted index
	local cTitlesKnown = 0
	local iMatch = 1
	local rgNew = {}
	
	for iID = 1, GetNumTitles(), 1 do
		if 1 == IsTitleKnown(iID) then
			cTitlesKnown = cTitlesKnown + 1
			local titleName = GetTitleName(iID)
			local fMatchMade, iRank = self:GetTitleRankFromSavedData(iMatch, iID)

			if (fMatchMade) then 
				iMatch = iMatch + 1 
			elseif (not self.db.profile.fFirstBoot) then
				self:Print(L["New title added! "] .. titleName)
				if (self.db.profile.fEnableHotness and self.db.profile.fNewHotness) then
					self.db.profile.idHotTitle = iID
				end
			end
			
			if (self.db.profile.idHotTitle == iID) then g_iHotTitle = cTitlesKnown end
						
			g_rgstrTitleNames[cTitlesKnown] = titleName
			g_rgstrHotTitleNames[cTitlesKnown] = titleName
			rgNew[cTitlesKnown] = { titleName, iID, iRank }
		end
	end

	if (#g_rgstrTitleNames < 1) then
		g_rgstrTitleNames[1] = L["no titles available"]
	end

	g_rgstrHotTitleNames[0] = L["no hot title"]

	
	-- now that we've read everything in, blow away old saved ranking with new list
	self.db.profile.rgTitleInfo = rgNew
	
	g_iTitleForRanking = 1
end


function Pokedex:GetTitleRankFromSavedData(iStart, iID)
	local cLimit = #self.db.profile.rgTitleInfo
	local index
	
	if (iStart <= cLimit) then
		for index = iStart, cLimit, 1 do
			if (self.db.profile.rgTitleInfo[index][TI_ID] == iID) then
				return true, self.db.profile.rgTitleInfo[index][TI_RANK]
			end
		end
		
		for index = 1, iStart - 1, 1 do
			if (self.db.profile.rgTitleInfo[index][TI_ID] == iID) then
				return true, self.db.profile.rgTitleInfo[index][TI_RANK]
			end
		end
	end
	
	return false, g_iDefaultRank
end


--=========================================================================--
-- Title options UI functions
--=========================================================================--

function Pokedex:FGetChangeTitleOnLoad(info) return self.db.profile.fChangeTitleOnLoad; end
function Pokedex:SetChangeTitleOnLoad(info, fNew) self.db.profile.fChangeTitleOnLoad = fNew; end

function Pokedex:GetTitleForRanking(info) return g_iTitleForRanking; end
function Pokedex:SetTitleForRanking(info, iNew) g_iTitleForRanking = iNew; end

function Pokedex:GetRankingForTitle(info)
	if (#self.db.profile.rgTitleInfo > 0) then
		return self.db.profile.rgTitleInfo[g_iTitleForRanking][TI_RANK]
	else
		return g_iDefaultRank
	end
end

function Pokedex:SetRankingForTitle(info, iNew)
	self.db.profile.rgTitleInfo[g_iTitleForRanking][TI_RANK] = iNew
end

function Pokedex:FDisableRankingForTitle(info) return not (#self.db.profile.rgTitleInfo > 0); end


function Pokedex:GetHotTitle(info) return g_iHotTitle; end
function Pokedex:SetHotTitle(info, iNew)
	g_iHotTitle = iNew
	if (iNew > 0) then
		self.db.profile.idHotTitle = self.db.profile.rgTitleInfo[iNew][TI_ID]
	else
		self.db.profile.idHotTitle = 0;
	end
end

function Pokedex:GetTitleHeat(info) return self.db.profile.iHotTitlePref; end
function Pokedex:SetTitleHeat(info, iNew) self.db.profile.iHotTitlePref = iNew; end
